#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dmzgraph.h"
#include "dmzgraph_internal.h"

#include <SDL/SDL.h>

SDL_Surface *primary = NULL;
SDL_Surface *surfaces[MAX_IMAGES];
int graph_last_error;
int graph_init_done = 0;

int dmz_graph_init(void* hwnd)
{
	int i, rv;
	
	if (graph_init_done) return DMZ_GRAPH_ALREADY_INIT;

	rv = SDL_Init(SDL_INIT_VIDEO);
	if (rv < 0)
		return DMZ_GRAPH_SDL_ERROR;
	
	for (i = 0; i < MAX_IMAGES; i++)
		surfaces[i] = NULL;
		
	graph_init_done = 1;
	return DMZ_GRAPH_SUCCESS;
}

int dmz_graph_kill(void)
{
	int i;

	if (!graph_init_done) return DMZ_GRAPH_NOT_INIT;

	for (i = 0; i < MAX_IMAGES; i++)
	{
		if (surfaces[i] != NULL)
			SDL_FreeSurface(surfaces[i]);
	}
	SDL_Quit();
	
	return DMZ_GRAPH_SUCCESS;
}

int dmz_graph_set_mode(int width, int height, int depth)
{
	Uint32 flags = SDL_HWSURFACE | SDL_DOUBLEBUF;

	if (!graph_init_done) return DMZ_GRAPH_NOT_INIT;
	
	primary = SDL_SetVideoMode(width, height, depth, flags);
	
	surfaces[0] = primary;
	
	if (primary == NULL)
		return DMZ_GRAPH_SDL_ERROR;
	
	return DMZ_GRAPH_SUCCESS;
}

int dmz_graph_flip(void)
{
	if (!graph_init_done) return DMZ_GRAPH_NOT_INIT;

	graph_last_error = SDL_Flip(primary);
	
	return DMZ_GRAPH_SUCCESS;
}

int dmz_graph_load_image(char *filename, int *width, int *height)
{
	int idx = get_available_surf();
	SDL_Surface *temp;
	
	if (!graph_init_done)
	{
		/* temp hack TODO to allow server to run without crashing */
		surfaces[idx] = (SDL_Surface *) 0xFFFFFFFF;
		return idx;
	}

	if (idx < 0) return DMZ_GRAPH_TOO_MANY;
	
	temp = SDL_LoadBMP(filename);
	surfaces[idx] = SDL_DisplayFormat(temp);
	
	if (surfaces[idx] == NULL)
		idx = DMZ_GRAPH_SDL_ERROR;
	else
	{
		if (width != NULL)
			*width = surfaces[idx]->w;
		if (height != NULL)
			*height = surfaces[idx]->h;
	}
	
	return idx;
}

int dmz_graph_create_virtual(int width, int height)
{
	int idx = get_available_surf();
	Uint32 flags = SDL_SWSURFACE;
	int depth = primary->format->BitsPerPixel;
	
	if (!graph_init_done) return DMZ_GRAPH_NOT_INIT;
	if (idx < 0) return DMZ_GRAPH_TOO_MANY;
	
	surfaces[idx] = SDL_CreateRGBSurface(flags, width, height, depth,
		primary->format->Rmask, primary->format->Gmask,
		primary->format->Bmask, primary->format->Amask);

	if (surfaces[idx] == NULL)
		idx =  DMZ_GRAPH_SDL_ERROR;

	return idx;
}

int dmz_graph_blit(int imgID, int src_x, int src_y, int src_w, int src_h,
			int dst_x, int dst_y, int transparent)
{
	return dmz_graph_virblit(imgID, 0, src_x, src_y, src_w, src_h,
		dst_x, dst_y, transparent);
}

int dmz_graph_virblit(int srcID, int dstID, int src_x, int src_y,
			int src_w, int src_h, int dst_x, int dst_y, int transparent)
{
	int i, j, rv;
	unsigned char *lpbSrc, *lpbDst;
	unsigned int sBpp, dBpp, sPitch, dPitch;

	if (!graph_init_done) return DMZ_GRAPH_NOT_INIT;
	
	rv = SDL_LockSurface(surfaces[srcID]);
	if (rv < 0)
		return DMZ_GRAPH_SDL_ERROR;

	rv = SDL_LockSurface(surfaces[dstID]);
	if (rv < 0)
	{
		SDL_UnlockSurface(surfaces[srcID]);
		return DMZ_GRAPH_SDL_ERROR;
	}
	
	sBpp = surfaces[srcID]->format->BytesPerPixel;
	dBpp = surfaces[dstID]->format->BytesPerPixel;
	
	sPitch = surfaces[srcID]->pitch;
	dPitch = surfaces[dstID]->pitch;

	lpbSrc = ((unsigned char *) surfaces[srcID]->pixels)
		+ src_y * surfaces[srcID]->pitch + src_x * sBpp;
	lpbDst = ((unsigned char *) surfaces[dstID]->pixels)
		+ dst_y * surfaces[dstID]->pitch + dst_x * dBpp;
	
	if (transparent < 0)
		for (i = 0; i < src_h; i++)
		{
			memcpy(lpbDst, lpbSrc, src_w * sBpp);
			
			lpbSrc += sPitch;
			lpbDst += dPitch;
		}
	
	else if (dBpp == 4)	/* 32-bit mode */
	for (i = 0; i < src_h; i++)
	{
		Uint32 *src = (Uint32 *) lpbSrc;
		Uint32 *dst = (Uint32 *) lpbDst;

		for (j = 0; j < src_w; j++)
			if (src[j] != transparent)
				dst[j] = src[j];
								
		lpbSrc += sPitch;
		lpbDst += dPitch;
	}
	else if (dBpp == 2) /* 16-bit mode */
	for (i = 0; i < src_h; i++)
	{
		Uint16 *src = (Uint16 *) lpbSrc;
		Uint16 *dst = (Uint16 *) lpbDst;

		for (j = 0; j < src_w; j++)
			if (src[j] != (Uint16) transparent)
				dst[j] = src[j];
								
		lpbSrc += sPitch;
		lpbDst += dPitch;
	}

	SDL_UnlockSurface(surfaces[srcID]);
	SDL_UnlockSurface(surfaces[dstID]);

	return DMZ_GRAPH_SUCCESS;
}

int dmz_graph_screenshot(char *filename)
{
	if (!graph_init_done) return DMZ_GRAPH_NOT_INIT;

	SDL_SaveBMP(primary, "screenshot.bmp");
	
	return DMZ_GRAPH_SUCCESS;
}

int dmz_graph_get_last_error(void)
{
	return graph_last_error;
}

int get_available_surf(void)
{
	int i;
	
	for (i = 0; i < MAX_IMAGES; i++)
		if (surfaces[i] == NULL)
			return i;
			
	return -1;
}
